# Copyright (c) Qotto, 2021

"""
Common integration utilities
----------------------------

Utility functions to integrate the eventy protocol in your workflow

Provides decorator factories to decorate functions and automatically extract tracing context
"""

import contextvars
import logging
from typing import Callable

from ..tracing.trace_generators import TraceGenerator
from ..tracing.trace_vars import TraceVar

logger = logging.getLogger(__name__)

__all__ = [
    'trace_decorator',
    'merge_decorators',
    'DecoratorType',
]

DecoratorType = Callable[..., Callable]


def merge_decorators(
    *decorators,
) -> DecoratorType:
    """
    Decorator factory merging multiple decorators, apply from left to right

    :param decorators: decorators to merge
    :return: a new decorator applying all decorators
    """

    if not decorators:
        def identity(func):
            return func

        return identity

    def decorator(func):
        def decorated_func(*args, **kwargs):
            inner_func = func
            nested_decorators = decorators
            for nested_decorator in nested_decorators:
                inner_func = nested_decorator(inner_func)

            return inner_func(*args, **kwargs)

        return decorated_func

    return decorator


def trace_decorator(
    trace_var: TraceVar,
    trace_generator: TraceGenerator,
    fetch: bool = True,
    override: bool = False,
) -> DecoratorType:
    """
    Decorator factory fetching trace from kwargs

    :param trace_var: TraceVar (correlation_id_var or request_id_var)
    :param trace_generator: TraceGenerator to generate trace id from function name func_name
    :param fetch: should the trace be fetched from decorated function kwargs?
    :param override: should the new trace override existing trace if present?
    :return: the configured decorator
    """

    def decorator(func):
        def decorated_func(*args, **kwargs):
            def inner():
                if trace_var.name in kwargs and fetch:
                    trace_id = kwargs.pop(trace_var.name)

                    if override or not trace_var.get():
                        logger.debug(f'Will set {trace_var.name}={trace_id} from {func.__name__} kwargs')
                        trace_var.set(trace_id)

                    else:
                        logger.debug(f'Will not use {trace_var.name}={trace_id} from {func.__name__} kwargs')

                elif trace_generator:
                    trace_id = trace_generator.generate_trace(
                        func_name=func.__name__,
                        func_args=args, func_kwargs=kwargs
                    )

                    if override or not trace_var.get():
                        logger.debug(f'Will set {trace_var.name}={trace_id} generated for {func.__name__}')
                        trace_var.set(trace_id)

                    else:
                        logger.debug(f'Will not use {trace_var.name}={trace_id} generated for {func.__name__}')

                return func(*args, **kwargs)

            context = contextvars.copy_context()
            result = context.run(inner)
            return result

        renamed_decorated_func = decorated_func
        renamed_decorated_func.__name__ = func.__name__
        return renamed_decorated_func

    return decorator



# Copyright (c) Qotto, 2021

"""
Trace variables definition

Define TraceVar for trace ids: correlation_id and request_id
"""

from contextvars import ContextVar

TraceVar = ContextVar

correlation_id_var: TraceVar = TraceVar("correlation_id", default='')
"""
correlation_id trace var

>>> from eventy.tracing import correlation_id_var
>>> correlation_id_var.set('my_operation_using_many_services')
>>> print(correlation_id_var.get())
'my_operation_using_many_services'
"""

request_id_var: TraceVar = TraceVar("request_id", default='')
"""
request_id trace var

>>> from eventy.tracing import request_id_var
>>> request_id_var.set('my_service_request')
>>> print(request_id_var.get())
'my_service_request'
"""



# Copyright (c) Qotto, 2021

"""
Trace IDs generation utilities
"""

import logging
from base64 import b64encode
from datetime import datetime, timezone
from secrets import token_urlsafe

logger = logging.getLogger(__name__)


class TraceGenerator:
    """
    Trace generator interface
    """

    def generate_trace(self, **kwargs):
        """
        Generate a trace id depending on keyword arguments

        :param kwargs: keyword arguments to define the trace
        :return: a new trace id
        """
        raise NotImplementedError


class NamespaceNameDateRandomTraceGenerator(TraceGenerator):
    def generate_trace(self, namespace: str = '', name: str = '', date: str = '', random: str = '', **kwargs):
        return f'{namespace}:{name}:{date}:{random}'


def _generate_encoded_date():
    ts_now = datetime.now(timezone.utc).timestamp()
    ts_2k = datetime(2000, 1, 1, tzinfo=timezone.utc).timestamp()
    ts2000res65536 = int(65536 * (ts_now - ts_2k)).to_bytes(6, 'big')
    encoded_date = b64encode(ts2000res65536, b'_-').decode('ascii')
    return encoded_date


def _generate_random(nbytes: int = 3):
    random = token_urlsafe(nbytes)
    return random


class NamespaceNameTraceGenerator(TraceGenerator):
    """
    Trace generator from a ``namespace`` and ``func_name``
    """

    def generate_trace(self, namespace=None, name=None, **kwargs):
        """
        Generate a trace depending on a function name in a namespace

        :param namespace: function name
        :param name: function name
        :param kwargs: other keyword args are ignored
        :return: a new trace id "prefix:func_name:encoded_date:random"
        """
        return NamespaceNameDateRandomTraceGenerator().generate_trace(
            namespace=namespace,
            name=name,
            date=_generate_encoded_date(),
            random=_generate_random(),
        )



# Copyright (c) Qotto, 2021

"""
Celery integration utilities

Utility functions to integrate the eventy protocol in celery apps

The celery module works best with celery optional dependencies installed ("pip install eventy[celery]")
"""

import logging
from contextvars import copy_context
from functools import wraps
from typing import Callable, List, Dict
from uuid import uuid4

from celery import shared_task as original_shared_task
from celery.execute import send_task as original_send_task
from celery.result import AsyncResult

from .common import DecoratorType
from ..tracing import correlation_id_var, request_id_var

logger = logging.getLogger(__name__)

__all__ = [
    'send_task',
    'shared_task',
    'with_namespace',
]


def send_task(name: str, args: List = [], kwargs: Dict = {}, **options) -> AsyncResult:
    """
    Modified version of celery.execute.send_task adding context in task kwargs

    options ar propagated as celery send_task named arguments and keyword arguments

    TODO: shouldn't we match argument names?
    ref: https://github.com/celery/celery/blob/494cc5d67452038c9b477d41cb2760b33ab4d5b8/celery/app/base.py#L704

    Usage::

        from eventy.integration.celery import send_task

        send_task(
            'service.task.name', [],
            {
                'param': 'value',
            }
        )

    Is equivalent to::

        from celery.execute import send_task
        from eventy.tracing import correlation_id_var

        send_task(
            'service.task.name', [],
            {
                'correlation_id': correlation_id_var.get(),
                'param': 'value',
            }
        )
    """
    if correlation_id_var.get():
        kwargs.update(correlation_id=correlation_id_var.get())
        logger.debug('Added correlation_id in task kwargs')
    else:
        logger.debug('No correlation_id in context to add in task kwargs')

    return original_send_task(name, args, kwargs, **options)


def shared_task(_func=None, **kwargs) -> DecoratorType:
    """
    Modified version of celery.shared_task fetching correlation id and generating request_id

    Usage::

        from eventy.integration.celery import shared_task

    Instead of::

        from celery import shared_task

    And then::

        @shared_task(options...)
        def my_task(param):
            pass


    :param _func: if used as a decorator with no args, decorated function will be first arg
    :param kwargs: optional ``namespace``, other options are propagated to celery's shared_task
    :return: new version of shared_task
    """

    def shared_task_decorator(function: Callable):

        @wraps(function)
        def decorated_shared_task(*func_args, **func_kwargs):
            if 'namespace' in kwargs:
                kwargs.update(name=f'{kwargs.pop("namespace")}.{function.__name__}')
            task_function = original_shared_task(**kwargs)(function)

            def inner():
                if 'correlation_id' in func_kwargs:
                    correlation_id_var.set(func_kwargs.pop('correlation_id'))
                else:
                    correlation_id_var.set(f'C:{function.__module__}.{function.__name__}:{uuid4()}')
                request_id_var.set(f'R:{function.__module__}.{function.__name__}:{uuid4()}')

                return task_function(*func_args, **func_kwargs)

            context = copy_context()
            result = context.run(inner)
            return result

        return decorated_shared_task

    if _func:
        # @shared_task
        # def f():
        return shared_task_decorator(_func)
    else:
        # @shared_task(param=value)
        # def f():
        return shared_task_decorator


def with_namespace(namespace: str, decorator: DecoratorType):
    """
    shared_task decorator factory adding namespace in task name

    Usage::

        from eventy.integration.celery import shared_task, with_namespace

        namespaced_shared_task = with_namespace('my_namespace', shared_task)

        @namespaced_shared_task(options...)
        def my_task(params...):
            pass

    :param namespace: namespace to configure the decorator with
    :param decorator: decorator to configure
    :return:
    """

    def decorator_with_namespace(_func=None, **kwargs):
        kwargs.update(namespace=namespace)
        return decorator(_func, **kwargs)

    return decorator_with_namespace



# Copyright (c) Qotto, 2021

class SerializationError(Exception):
    """
    Base for all serialization errors
    """


class UnknownRecordTypeError(SerializationError):
    """
    Record type unknown (not EVENT, REQUEST, or RESPONSE)
    """

    def __init__(self, record_type: str):
        super().__init__(f'Unknown record type {record_type}')



# Copyright (c) Qotto, 2021

"""
Include support for context bound trace ids

* :obj:`trace_vars.correlation_id_var` a Correlation ID to correlate multiple services
  involved in the same business operation
* :obj:`trace_vars.request_id_var` a Request ID to identify a single request on a service

Include support to generate trace ids

* :class:`trace_generators.FuncNameTraceGenerator` trace id generator from a function name
"""

from .trace_generators import TraceGenerator
from .trace_vars import correlation_id_var, request_id_var

__all__ = [
    'correlation_id_var',
    'request_id_var',
    'TraceGenerator',
]
